/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * adc.c -- A/D interface source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/

#include "stdinc.h"
#include "counter.h"
#include "pci703def.h"
#include "adc.h"

void adc_init( pADC_t padc, pCNT_t pcnt, void *datafifo, void *chanfifo,
		void *control, void *status)
{
	pADCBuffer_t pbuf = &padc->buffer;
	spin_lock_init(&padc->lock);
	spin_lock_init(&pbuf->lock);
	spin_lock(&padc->lock);
	padc->datafifo = datafifo;
	padc->chanfifo = chanfifo;
	padc->control = control;
	padc->status = status;
	padc->pcnt = pcnt;
	padc->reg.clksrc = ADC_CLKSRC_SOFTWARE;
	padc->reg.trigsrc = ADC_TRIGSRC_INTERNAL;
	padc->reg.trigmode = ADC_TRIGMDE_POS;
	padc->frequency = ADC_FREQUENCY_MAX;
	padc->reg.clksrc = ADC_CLKSRC_INTERNAL;
	padc->reg.status = ADC_STATUS_IDLE;
	padc->reg.softclk = 0;
	padc->reg.eolints = ADC_IRQ_DISABLED;
	padc->testmux = 0;
	padc->chanlstln = 0;
	padc->datamode = ADC_DATA_MODE_VOLT;
	padc->waterlevel = ADC_FIFO_HALF;
	spin_unlock(&padc->lock);
	adc_buffrst(&padc->buffer);
	pbuf->data=(unsigned short *)vmalloc(ADC_BUFFER_SIZE*sizeof(unsigned short));
	if(pbuf->data == NULL) {
		drv_err("%s: Failed to allocate A/D circular buffer.", DEVICE);
		pbuf->size = 0;
	}
	else {
		drv_info("%s: Allocated %d kb for A/D circular buffer.", DEVICE, (ADC_BUFFER_SIZE * sizeof(unsigned short)) / 1024);
		pbuf->size = ADC_BUFFER_SIZE;
	}
	adc_fiforst(padc);
	if(adc_fifotest(padc))
		drv_err("%s: A/D FIFO hardware error!\n", DEVICE);
}

void adc_uninit( pADC_t padc )
{
	pADCBuffer_t pbuf = &padc->buffer;
	adc_fullstop(padc);
	if(pbuf->size > 0)
		vfree(pbuf->data);
}

long adc_isbusy( pADC_t padc )
{
	long status;
	spin_lock(&padc->lock);
	status=(long)padc->reg.status;
	spin_unlock(&padc->lock);
	return status;
}

long adc_fifotest( pADC_t padc )
{
	int i;
	unsigned long depth = 4096;
	long eerr = 0;
	for(i=0; i<depth; i++)
		writel(i, padc->datafifo);
	for(i=0; i<depth; i++)
		if(readl(padc->datafifo) != i) eerr |= 1;
	for(i=0; i<depth; i++)
		writel(16383, padc->datafifo);
	for(i=0; i<depth; i++)
		if(readl(padc->datafifo) != 16383)
			eerr |= 2;
	for(i=0; i<depth; i++)
		writel(0, padc->datafifo);
	for(i=0; i<depth; i++)
		if(readl(padc->datafifo) != 0)
			eerr |= 4;
	for(i=depth; i>0; i--)
		writel(i, padc->datafifo);
	for(i=depth; i>0; i--)
		if(readl(padc->datafifo) != i)
			eerr |= 8;
	return eerr;
}

void adc_start( pADC_t padc )
{
	spin_lock(&padc->lock);
	padc->reg.clksrc = padc->clksrc;
	padc->reg.trigsrc = padc->trigsrc;
	padc->reg.status = ADC_STATUS_BUSY;
	padc->reg.eolints = ADC_IRQ_ENABLED;
	adc_ctrlupdate(padc);
	spin_unlock(&padc->lock);
}

void adc_stop( pADC_t padc )
{
	padc->reg.status = ADC_STATUS_BUSY;
	spin_lock(&padc->lock);
	padc->reg.clksrc = ADC_CLKSRC_SOFTWARE;
	padc->reg.eolints = ADC_IRQ_DISABLED;
	adc_ctrlupdate(padc);
	padc->reg.status = ADC_STATUS_IDLE;
	spin_unlock(&padc->lock);
}

void adc_fullstop( pADC_t padc )
{
	adc_stop(padc);
}

long adc_setclock( pADC_t padc, long clksrc, long trigsrc )
{
	if((clksrc>2) || (clksrc<1))
		return -EINVAL;
	if((trigsrc>2) || (trigsrc<0))
		return -EINVAL;
	spin_lock(&padc->lock);
	padc->reg.clksrc = clksrc;
	padc->reg.trigsrc = trigsrc;
	spin_unlock(&padc->lock);
	return 0;
}

long adc_setdatamode( pADC_t padc, unsigned char mode )
{
	if(mode > ADC_DATA_MODE_VOLT) mode = ADC_DATA_MODE_VOLT;
	padc->datamode = mode;
	return 0;
}

void adc_ctrlupdate( pADC_t padc )
{
	padc->reg.lastval =
			(
			(padc->reg.eolints<<12) |
			(padc->reg.softclk<<10) |
			(padc->reg.trigmode<<5) |
			(padc->reg.trigsrc<<3) |
			(padc->reg.clksrc<<1) |
			(padc->reg.status)
			)&0xffff;
	writel(padc->reg.lastval, padc->control);
}

void adc_fiforst( pADC_t padc )
{
	padc->reg.status = ADC_STATUS_BUSY;
	adc_ctrlupdate(padc);
	padc->reg.status = ADC_STATUS_IDLE;
	adc_ctrlupdate(padc);
	udelay(10);
}

long adc_fifoempty( pADC_t padc )
{
	if((readl(padc->status) & 2) == 2)
		return true;
	return false;
}

void adc_softclock( pADC_t padc )
{
	spin_lock(&padc->lock);
	padc->reg.softclk = 1;
	adc_ctrlupdate(padc);
	udelay(10);
	padc->reg.softclk = 0;
	adc_ctrlupdate(padc);
	spin_unlock(&padc->lock);
}

void adc_progchanlist( pADC_t padc )
{
	unsigned char bip, grp, mux, pairs, ref, chan, gain, range;
	unsigned long pos, value;
	for(pos=0; pos<padc->chanlstln; pos++)
	{
		chan = padc->chanlst[pos].channel;
		gain = padc->chanlst[pos].gain;
		range = padc->chanlst[pos].range;
		ref = padc->chanlst[pos].reference;
		bip = adc_isbipolar(range);
		encodechannel(chan, range, gain, &mux, &grp, &pairs);
		value= ( (gain<<11) |
			(3<<9) |
			(bip<<8) |
			(ref<<7) |
			(pairs<<5) |
			(grp<<2) |
			(mux) )&0x3fff;
		writel(value, padc->chanfifo);
	}
}

void encodechannel( unsigned char chan, unsigned char range, unsigned char gain,
		    unsigned char *mux, unsigned char *group, unsigned char *pairs )
{
	const unsigned char ADC_GRP_TAB[ADC_NUM_CHAN64] = {
				0,0,0,0,1,1,1,1,0,0,0,0,1,1,1,1,
				2,2,2,2,3,3,3,3,2,2,2,2,3,3,3,3,
				4,4,4,4,5,5,5,5,4,4,4,4,5,5,5,5,
				6,6,6,6,7,7,7,7,6,6,6,6,7,7,7,7 };
	const unsigned char ADC_MUX_TAB[ADC_NUM_CHAN64] = {
				0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3,
				0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3,
				0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3,
				0,1,2,3,0,1,2,3,0,1,2,3,0,1,2,3 };
	const unsigned char ADC_PAIR_TAB[ADC_NUM_CHAN64] ={
				1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
				1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
				1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2,
				1,1,1,1,1,1,1,1,2,2,2,2,2,2,2,2 };
	unsigned char diff = adc_isdifferential(range);
	*mux = ADC_MUX_TAB[chan];
	*group = ADC_GRP_TAB[chan];
	*pairs = 0;
	if(diff == 0) *pairs = ADC_PAIR_TAB[chan];
}

long adc_isdiffchan( unsigned char chan )
{
	if((chan%16)<8)
		return 1;
	return 0;
}

unsigned char adc_getdiffref( unsigned char chan )
{
	return (chan+8);
}

void adc_buffrst( pADCBuffer_t pbuf )
{
	spin_lock(&pbuf->lock);
	pbuf->samples = 0;
	pbuf->head = 0;
	pbuf->tail = 0;
	pbuf->wrap = 0;
	pbuf->overrun = 0;
	spin_unlock(&pbuf->lock);
}

void adc_reset( pADC_t padc )
{
	adc_fiforst(padc);
	adc_buffrst(&padc->buffer);
	padc->chanidx = 0;
}

unsigned long adc_setfrequency( pADC_t padc, unsigned long value )
{
	unsigned long  val;
	spin_lock(&padc->lock);
	if(padc->clksrc == ADC_CLKSRC_INTERNAL) {
		if(value > ADC_FREQUENCY_MAX)
			value = ADC_FREQUENCY_MAX;
		padc->frequency = 0;
		cnt_disable(padc->pcnt);
		if(value > 1000) {
			cnt_configure(padc->pcnt, 0, 0, 0);
			val=20000000 / ((unsigned long )value);
			padc->frequency = 20000000 / val;
		}
		else {
			cnt_configure(padc->pcnt, 1, 0, 0);
			val=100000 / ((unsigned long )value);
			padc->frequency = 100000 / val;
		}
	}
	else {
		value &= 0xffff;
		val = value;
	}
	cnt_write(padc->pcnt, val);
	cnt_enable(padc->pcnt);
	spin_unlock(&padc->lock);
	return padc->frequency;
}

long adc_getsinglevalue( pADC_t padc, long* buffer )
{
	int range, gain;	
	unsigned short raw;	
	gain = padc->chanlst[0].gain;
	range = padc->chanlst[0].range;
	if(readl(padc->control)&0x00000002) return -1;
	raw = readl(padc->datafifo);
	*buffer = adc_bintovolt(raw, range, gain);	
	return 0;
}

void adc_irqenable( pADC_t padc )
{
	spin_lock(&padc->lock);
	padc->reg.eolints = ADC_IRQ_ENABLED;
	adc_ctrlupdate(padc);
	spin_unlock(&padc->lock);
}

void adc_irqdisable( pADC_t padc )
{
	spin_lock(&padc->lock);
	padc->reg.eolints = ADC_IRQ_DISABLED;
	adc_ctrlupdate(padc);
	spin_unlock(&padc->lock);
}

long adc_settrigmode( pADC_t padc, long mode )
{
	if(mode > ADC_TRIGMDE_FALL)
		return -EINVAL;
	spin_lock(&padc->lock);
	padc->reg.trigmode = mode;
	adc_ctrlupdate(padc);
	spin_unlock(&padc->lock);
	return 0;
}

long adc_getwaterlevel( pADC_t padc )
{
	long value;
	spin_lock(&padc->lock);
	value = padc->waterlevel;
	spin_unlock(&padc->lock);
	return value;
}

long adc_isbipolar( long range )
{
	if((range == ADC_RANGE_BIPSE) || (range == ADC_RANGE_BIPDIF))
		return 1;
	return 0;
}

long adc_isdifferential( long range )
{
	if((range == ADC_RANGE_UNIDIF) || (range == ADC_RANGE_BIPDIF))
		return 1;
	return 0;
}

unsigned long adc_getunread( pADCBuffer_t pbuf )
{
	unsigned long samples, flags;
	spin_lock_irqsave(&pbuf->lock, flags);
	samples = pbuf->samples;
	spin_unlock_irqrestore(&pbuf->lock, flags);
	return samples;
}

long adc_bintovolt( unsigned short bin,int range, int gain )
{
	const int MULTBIP[9] = { 4,2,1,4,2,1,4,2,1 };
	const int MULTUNI[9] = { 2,2,1,4,2,1,4,2,1 };
	const int DIV[9] = { 1,1,1,10,10,10,100,100,100 };
	long uv, sign =1 ;
	if((range == 1) || (range == 3)) {
		if(bin >= 0x2000) {
			bin = ~bin;
			bin &= 0x1fff;
			sign = -1;
		}
		uv = bin * 305;
		uv += (bin * 175) / 1000;
		uv *= MULTBIP[gain] * sign;
	}
	else {
		if(bin >= 0x2000) {
			bin &= 0x1fff;
			sign = 0;
		}
		uv = (bin + 8192 * sign) * 305;
		uv += ((bin + 8192 * sign) * 175) / 1000;
		uv *= MULTUNI[gain];
	}
	uv /= DIV[gain];
	return uv;
}

long adc_dmatobuf( pADCBuffer_t pbuf, unsigned long* dmabuf,
		   unsigned short waterlevel )
{
	int eerr = 0;
	unsigned long flags, samples, i;
	spin_lock_irqsave(&pbuf->lock, flags);
	samples = waterlevel;	

	if(samples > (pbuf->size - pbuf->samples)) {
		drv_err("%s: A/D buffer overrun.", DEVICE);
		samples = pbuf->size - pbuf->samples;
		eerr = -EINVAL;
		pbuf->overrun = 1;
	}

	for(i=0; i<samples; i++) {
		pbuf->data[pbuf->head++] = (unsigned short)(dmabuf[i] & 0x3fff);
		if(pbuf->head == pbuf->size) {
			pbuf->head = 0;
			pbuf->wrap++;
		}
	}
	pbuf->samples += samples;
	spin_unlock_irqrestore(&pbuf->lock, flags);
	return eerr;
}

unsigned long adc_getdataraw( pADCBuffer_t pbuf, unsigned short *buffer,
			      unsigned long samples )
{
	unsigned long flags, copied;
	spin_lock_irqsave(&pbuf->lock, flags);
	if(samples > pbuf->samples)
		samples = pbuf->samples;
	copied = samples;
	if(samples > ((pbuf->size - 1) - pbuf->tail)) {
		memcpy( buffer,
			&pbuf->data[pbuf->tail],
			((pbuf->size-1)-pbuf->tail)*sizeof(unsigned short) );
		samples -= ((pbuf->size-1) - pbuf->tail);
		pbuf->samples += ((pbuf->size - 1) - pbuf->tail);
		pbuf->tail = 0;
	}
	memcpy( buffer, &pbuf->data[pbuf->tail], samples*sizeof(unsigned short));
	pbuf->tail += samples;
	pbuf->samples -= samples;
	spin_unlock_irqrestore(&pbuf->lock, flags);
	return copied;
}

unsigned long adc_getdata( pADC_t padc, long *buffer,
			   unsigned long samples )
{
	pADCBuffer_t pbuf = &padc->buffer;
	unsigned long flags, i;
	int range, gain;

	spin_lock_irqsave(&pbuf->lock, flags);
	if(samples > pbuf->samples)
		samples = pbuf->samples;
	for(i=0; i<samples; i++)
	{
		range = padc->chanlst[padc->chanidx].range;
		gain = padc->chanlst[padc->chanidx].gain;
		buffer[i] = adc_bintovolt(pbuf->data[pbuf->tail], range, gain);
		if(++padc->chanidx == padc->chanlstln)
			padc->chanidx = 0;
		if(++pbuf->tail == pbuf->size)
			pbuf->tail = 0;
	}
	pbuf->samples -= samples;
	spin_unlock_irqrestore(&pbuf->lock, flags);
	return samples;
}

/**
 * End of File
 */
