/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * dac.c -- D/A interface source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/

#include "stdinc.h"
#include "counter.h"
#include "dac.h"

void dac_init( pDAC_t pdac, void *control, void *datafifo, pCNT_t pcounter )
{
	spin_lock_init(&pdac->lock);
	spin_lock(&pdac->lock);
	pdac->ctrl = control;
	pdac->data = datafifo;
	pdac->status = DAC_STATUS_IDLE;
	pdac->pcnt = pcounter;
	dac_update(pdac);
	dac_progio(pdac, 0);
	spin_unlock(&pdac->lock);
}

void dac_uninit( pDAC_t pdac )
{
	dac_disable(pdac);
}

void dac_update( pDAC_t pdac )
{
	pdac->reg.lastvalue =
			(pdac->reg.fifohalf<<5) |
			(pdac->reg.fifoempty) |
			(pdac->reg. mode);
  writel(pdac->reg.lastvalue, pdac->ctrl);
}

unsigned short dac_volttobin( long uv )
{
	unsigned short bin;
	if(uv < 0) {
		uv = -uv;
		bin = uv / 1221;
		bin = 8191 - bin;
	}
	else {
		bin = uv / 1221;
		bin |= 0x2000;
	}
	if(bin > 16383)
		bin = 16383;
	return bin;
}

void dac_progio( pDAC_t pdac, long uv )
{
	pdac->reg.mode = DAC_MODE_PIO;
	dac_update(pdac);
	dac_writefifo(pdac, dac_volttobin(uv));
}

long dac_isbusy( pDAC_t pdac )
{
	long status;
	spin_lock(&pdac->lock);
	status = pdac->status;
	spin_unlock(&pdac->lock);
	return status;
}

void dac_pause( pDAC_t pdac )
{
	cnt_disable(pdac->pcnt);
}

void dac_reset( pDAC_t pdac )
{
	spin_lock(&pdac->lock);
	pdac->reg.mode = DAC_MODE_PIO;
	dac_update(pdac);
	pdac->reg.mode = DAC_MODE_DISABLED;
	dac_update(pdac);
	spin_unlock(&pdac->lock);
}

long dac_adddata( pDAC_t pdac, void *buffer, unsigned long length )
{
	int i;
	long *pbuf;
	if(length > DAC_FIFO_SIZE)
		return -EINVAL;
	pbuf = (long *)kmalloc(length*sizeof(long),GFP_KERNEL);
	if(pbuf == NULL)
		return -ENOMEM;
	if(copy_from_user(pbuf, buffer, length))
		return -ENOMEM;
	for(i=0; i<length; i++)
		writel(dac_volttobin(pbuf[i]), pdac->data);
	kfree(pbuf);
	return 0;
}

void dac_setmode( pDAC_t pdac, unsigned char mode )
{
	spin_lock(&pdac->lock);
	pdac->runmode = mode;
	spin_unlock(&pdac->lock);
}

void dac_enable( pDAC_t pdac )
{
	spin_lock(&pdac->lock);
	cnt_enable(pdac->pcnt);
	pdac->reg.mode = pdac->runmode;
	dac_update(pdac);
	spin_unlock(&pdac->lock);
}

void dac_disable( pDAC_t pdac )
{
	spin_lock(&pdac->lock);
	cnt_disable(pdac->pcnt);
	pdac->reg.mode = DAC_MODE_DISABLED;
	dac_update(pdac);
	spin_unlock(&pdac->lock);
}

void dac_pattern( pDAC_t pdac )
{
	spin_lock(&pdac->lock);
	cnt_enable(pdac->pcnt);
	pdac->reg.mode = DAC_MODE_PATTERN;
	dac_update(pdac);
	spin_unlock(&pdac->lock);
}

void dac_writefifo( pDAC_t pdac, unsigned short bin )
{
	writel(bin,pdac->data);
}

long dac_configure( pDAC_t pdac, long freq, long clksrc, long gatesrc )
{
	if(clksrc > 2)
		return -EINVAL;
	cnt_configure(pdac->pcnt, clksrc, CNT_MODE_PULSE, 0);
	cnt_write(pdac->pcnt, freq);
	return 0;
}

void dac_continue( pDAC_t pdac )
{
	cnt_enable(pdac->pcnt);
}

void dac_writebin(pDAC_t pdac, unsigned short bin)
{
	spin_lock(&pdac->lock);
	pdac->reg.mode = DAC_MODE_PIO;
	dac_writefifo(pdac, bin);
	spin_unlock(&pdac->lock);
}

/**
 * End of File
 */
