/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * pci703.c -- main source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/

#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
# error "This driver needs kernel 2.6.9 and later"
#endif
#include <linux/config.h>
#ifndef CONFIG_PCI
# error "This driver needs PCI support"
#endif
#include "stdinc.h"
#include "pci703def.h"
#include "eeprom.h"
#include "counter.h"
#include "adc.h"
#include "dac.h"
#include "caldac.h"
#include "isr.h"
#include "pci703.h"
#include "ioctl.h"
#include "version.h"

#define DRIVER_AUTHOR "Johan Russouw, johan@eagle.co.za & Jan Zwiegers, \
			jan@eagle.co.za"
#define DRIVER_DESC "PCI703 Interface Driver (C) 2004 - 2005 Eagle Technology"

static int dev_read( struct file *file, char *buf, size_t length,
			loff_t * offset)
{
	pBoardDevice_t pdev;
	pADC_t padc;
	unsigned long samples, tocopy, copied, transfered, size;
	int eerr = 0;

	pdev = (pBoardDevice_t)file->private_data;
	padc = &pdev->adc;
	size = sizeof(unsigned short);
	if(padc->datamode == ADC_DATA_MODE_VOLT)
		size = sizeof(long);

	samples = length / size;
	transfered = 0;
	do {
		tocopy = samples;
		if(tocopy > (pdev->dblbuffer.size / size))
			tocopy = pdev->dblbuffer.size / size;
		if(padc->datamode != ADC_DATA_MODE_VOLT)
			copied = adc_getdataraw( &padc->buffer,
				(unsigned short *)pdev->dblbuffer.data, tocopy);
		else
			copied = adc_getdata(padc, (long *)pdev->dblbuffer.data, tocopy);
		samples -= copied;
		eerr = tocopy - copied;
		do {
			tocopy = copied;
			copied = copy_to_user( (void *)(buf + transfered * size),
					(void *)pdev->dblbuffer.data, tocopy * size);
			if(copied)
				copied /= size;
			transfered += tocopy - copied;
		}while(copied > 0);
	}while(samples > 0 || eerr);
	transfered -= samples;
	if(eerr)
		return -EINVAL;
	return transfered * size;
}

static int dev_write(struct file *file, const char *buf, size_t length,
			loff_t * offset)
{
	return -EINVAL;
}

static void dev_kref_release( struct kref *kref )
{
	struct BoardDevice_t *dev = NULL;
	dev = container_of(kref, struct BoardDevice_t, kref);
	kfree(dev);
}

static int dev_open( struct inode *inode, struct file *pfile )
{
	pBoardDevice_t pdev;
	pdev = (pBoardDevice_t)pfile->private_data;
	spin_lock(&pdev->lock);
	spin_unlock(&pdev->lock);
	pfile->f_pos=0;
	kref_get(&pdev->kref);
	return 0;
}

static int dev_release( struct inode *inode, struct file *pfile )
{
	pBoardDevice_t pdev = (pBoardDevice_t)pfile->private_data;
	kref_put(&pdev->kref, dev_kref_release);
	return 0;
}

static int dev_ioctl( struct inode *inode, struct file *pfile,
		      unsigned int cmd, unsigned long arg )
{
	pBoardDevice_t pdev = (pBoardDevice_t)pfile->private_data;
	int eerr = 0;
	int size = _IOC_SIZE(cmd);
	if(_IOC_TYPE(cmd) != DEV_IOC_MAGIC)
		return -EINVAL;
	if(_IOC_NR(cmd) > DEV_IOC_MAXNR)
		return -EINVAL;
	if(_IOC_DIR(cmd) & _IOC_READ)
		eerr = verify_area(VERIFY_WRITE, (void *)arg, size);
	else if(_IOC_DIR(cmd) & _IOC_WRITE)
		eerr = verify_area(VERIFY_READ, (void *)arg, size);
	if(eerr)
		return eerr;

	switch(cmd) {
	case DEV_IOC_QUERY:
		eerr = ioctl_query(pdev, arg);
		break;
	case DEV_IOC_DIOWRITE:
		eerr = ioctl_diowrite(pdev, arg);
		break;
	case DEV_IOC_DIOREAD:
		eerr = ioctl_dioread(pdev, arg);
		break;
	case DEV_IOC_MIOCFG:
		eerr = ioctl_miocfg(pdev, arg);
		break;
	case DEV_IOC_CTREAD:
		eerr = ioctl_ctread(pdev, arg);
		break;
	case DEV_IOC_CTWRITE:
		eerr = ioctl_ctwrite(pdev, arg);
		break;
	case DEV_IOC_CTCFG:
		eerr = ioctl_ctcfg(pdev, arg);
		break;
	case DEV_IOC_CTSOFTGATE:
		eerr = ioctl_ctsoftgate(pdev, arg);
		break;
	case DEV_IOC_DAWRITE:
		eerr = ioctl_dawrite(pdev,arg);
		break;
	case DEV_IOC_DACONTROL:
		eerr = ioctl_dacontrol(pdev,arg);
		break;
	case DEV_IOC_DABLOCK:
		eerr = ioctl_dablock(pdev,arg);
		break;
	case DEV_IOC_DACFG:
		eerr = ioctl_dacfg(pdev,arg);
		break;
	case DEV_IOC_ADSINGLE:
		eerr = ioctl_adsingle(pdev,arg);
		break;
	case DEV_IOC_ADCFG:
		eerr = ioctl_adconfig(pdev,arg);
		break;
	case DEV_IOC_ADCLKSRC:
		eerr = ioctl_adclksrc(pdev, arg);
		break;
	case DEV_IOC_ADCHLST:
		eerr = ioctl_adchlst(pdev, arg);
		break;
	case DEV_IOC_ADFREQ:
		eerr = ioctl_adfreq(pdev, arg);
		break;
	case DEV_IOC_ADSTART:
		eerr = ioctl_adstart(pdev, arg);
		break;
	case DEV_IOC_ADSTOP:
		eerr = ioctl_adstop(pdev, arg);
		break;
	default:
		eerr = -EINVAL;
		break;
	}
	return eerr;
}

static struct file_operations device_fops = {
	.owner = THIS_MODULE,
	.open = dev_open,
	.release = dev_release,
	.ioctl = dev_ioctl,
	.read = dev_read,
	.write = dev_write,
};

void dev_inittype( pBoardDevice_t pdev )
{
	switch(pdev->devinfo.boardtype) {
		case PCI70316:
			pdev->devinfo.adcchan = 16;
			pdev->devinfo.dacchan = 0;
			break;
		case PCI70316A:
			pdev->devinfo.adcchan = 16;
			pdev->devinfo.dacchan = 2;
			break;
		case PCI70332:
			pdev->devinfo.adcchan = 32;
			pdev->devinfo.dacchan = 0;
			break;
		case PCI70332A:
			pdev->devinfo.adcchan = 32;
			pdev->devinfo.dacchan = 2;
			break;
		case PCI70364:
			pdev->devinfo.adcchan = 64;
			pdev->devinfo.dacchan = 0;
			break;
		case PCI70364A:
			pdev->devinfo.adcchan = 64;
			pdev->devinfo.dacchan = 2;
			break;
		default:
			pdev->devinfo.adcchan = 0;
			pdev->devinfo.dacchan = 0;
			break;
	}
	pdev->devinfo.diochan = 18;
	pdev->devinfo.cntchan = 0;
	pdev->devinfo.miochan = 10;
}


void dev_readeeprom( pBoardDevice_t pdev )
{
	long i;
	unsigned short uval;
	pdev->devinfo.serialnumber = ee_read(&pdev->eeprom, 31);
	pdev->devinfo.serialnumber |= ee_read(&pdev->eeprom, 30)<<16;
	pdev->devinfo.year = ee_read(&pdev->eeprom, 32);
	pdev->devinfo.month = ee_read(&pdev->eeprom, 33)&0xff;
	pdev->devinfo.day = (ee_read(&pdev->eeprom, 33)>>8)&0xff;
	pdev->devinfo.boardtype = ee_read(&pdev->eeprom, 34);
	pdev->devinfo.revision = ee_read(&pdev->eeprom, 35);
	caldac_enable(&pdev->caldac);
	for(i=0; i<8; i++) {
		uval = ee_read(&pdev->eeprom, i);
		caldac_write(&pdev->caldac, i, uval);
		uval=ee_read(&pdev->eeprom, i + 10);
		writel(uval, pdev->ptrmem0 + PCI703_REG_GAIN_CAL_BUF + (i * 4));
	}
}

int dev_hwinit( pBoardDevice_t pdev )
{

	irq_init( &pdev->irq,
		   pdev->ptrmem1 + PCI703_REG_GLOBAL_STATUS,
		   pdev->ptrmem0 + PCI703_REG_DMA_CFG,
		   pdev->ptrmem0 + PCI703_REG_AD_FIFO_DATA,
		   &pdev->adc.buffer);
	caldac_init( &pdev->caldac, pdev->ptrmem1 + PCI703_REG_GLOBAL_STATUS);
	ee_init(&pdev->eeprom,pdev->ptrmem1 + PCI703_REG_GLOBAL_STATUS);
	dev_readeeprom(pdev);
	if( (pdev->devinfo.serialnumber < 1000000001) ||
		(pdev->devinfo.serialnumber == 0xffffffff)) {
		drv_err( "%s: Invalid Serial Number (%ld).",
		     DEVICE, pdev->devinfo.serialnumber);
		return -ENODEV;
	}
	if( (pdev->devinfo.boardtype <= PCI70316) ||
		(pdev->devinfo.boardtype > PCI70332A)) {
		drv_err( "%s: Invalid Board Type (%d).",
		     DEVICE, pdev->devinfo.boardtype);
		return -ENODEV;
	}

	drv_info( "%s: Added new device (serial number #%ld, type #%d).",
	      DEVICE, pdev->devinfo.serialnumber, pdev->devinfo.boardtype);

	dev_inittype(pdev);
	cnt_init( &pdev->cnt[0], 0, pdev->ptrmem1 + PCI703_REG_CNTR_CNTRL0,
		  pdev->ptrmem1 + PCI703_REG_CNTR_DIVR0);
	cnt_init( &pdev->cnt[1], 1, pdev->ptrmem1 + PCI703_REG_CNTR_CNTRL1,
		   pdev->ptrmem1 + PCI703_REG_CNTR_DIVR1);
	cnt_init( &pdev->cnt[2], 2, pdev->ptrmem1 + PCI703_REG_CNTR_CNTRL2,
		   pdev->ptrmem1 + PCI703_REG_CNTR_DIVR2);
	adc_init( &pdev->adc, &pdev->cnt[2],
		  pdev->ptrmem0 + PCI703_REG_AD_FIFO_DATA,
		  pdev->ptrmem0 + PCI703_REG_AD_CHAN_LIST,
		  pdev->ptrmem1 + PCI703_REG_AD_CNTRL,
		  pdev->ptrmem1 + PCI703_REG_GLOBAL_STATUS);
	pdev->adc.channels = pdev->devinfo.adcchan;
	pdev->dio[0].setup = 0;
	pdev->dio[0].value = 0;
	pdev->dio[1].setup = 0;
	pdev->dio[1].value = 0;
	pdev->irq.waterlevel = DEFAULT_WATERLEVEL;

	if(pdev->devinfo.dacchan > 0) {
	dac_init( &pdev->dac[0],
		  pdev->ptrmem1 + PCI703_REG_DAC_CNTRL0,
		  pdev->ptrmem0 + PCI703_REG_DAC_FIFO0,
		  &pdev->cnt[0]);
	dac_init( &pdev->dac[1],
		  pdev->ptrmem1 + PCI703_REG_DAC_CNTRL1,
		  pdev->ptrmem0 + PCI703_REG_DAC_FIFO1,
		  &pdev->cnt[1]);
	}
	
	irq_enable(&pdev->irq);
	pdev->dblbuffer.data = kmalloc(DOUBLE_BUFFER_SIZE, GFP_KERNEL);
	if(pdev->dblbuffer.data == NULL)
		pdev->dblbuffer.size = 0;
	else
		pdev->dblbuffer.size = DOUBLE_BUFFER_SIZE;
	drv_info( "%s: Allocated double buffer of %d bytes.",
	      DEVICE, pdev->dblbuffer.size);
	return 0;
}

void dev_hwuninit( pBoardDevice_t pdev )
{
	adc_uninit(&pdev->adc);
	dac_uninit(&pdev->dac[0]);
	dac_uninit(&pdev->dac[1]);
	irq_uninit(&pdev->irq);
}

int dev_alloc( struct pci_dev *pcidev, pBoardDevice_t pdev )
{
	spin_lock_init(&pdev->lock);
	spin_lock(&pdev->lock);
	pdev->pci_device = pcidev;
	pdev->irq.pci_device = pcidev;

	pdev->ptrmem0 = ioremap_nocache( pci_resource_start(pcidev, 0),
					 pci_resource_len(pcidev, 0));
	if(!pdev->ptrmem0) {
		spin_unlock(&pdev->lock);
		drv_err("%s: Unable to remap BAR 0.", DEVICE);
		return -ENOMEM;
	}

	pdev->ptrmem1=ioremap_nocache( pci_resource_start(pcidev, 1),
				       pci_resource_len(pcidev, 1));

	if(!pdev->ptrmem1) {
		spin_unlock(&pdev->lock);
		drv_err("%s: Unable to remap BAR 1.", DEVICE);
		iounmap(pdev->ptrmem0);
		return -ENOMEM;
	}

	pci_set_master(pcidev);

	if(dev_hwinit(pdev)) {
		spin_unlock(&pdev->lock);
		drv_err("%s: Failed to initialize hardware.", DEVICE);
		return -ENODEV;
	}
	spin_unlock(&pdev->lock);
	return 0;
}

void dev_dealloc( pBoardDevice_t pdev )
{
	spin_lock(&pdev->lock);
	dev_hwuninit(pdev);
	iounmap(pdev->ptrmem0);
	iounmap(pdev->ptrmem1);
	spin_unlock(&pdev->lock);
}

static ssize_t dev_show_serial( struct device *device, char *buf )
{
	char s[20];
	pBoardDevice_t pbd = (pBoardDevice_t)device->driver_data;
	sprintf(s,"%ld\n", pbd->devinfo.serialnumber);
	strcpy(buf, s);
	return strlen(s) + 1;
}
static DEVICE_ATTR(serial, S_IRUGO, dev_show_serial, NULL);

static int __devinit dev_probe( struct pci_dev *dev,
				const struct pci_device_id *id )
{
	pBoardDevice_t pbd;

	drv_info( "%s: Probing device: vendor = %x, device = %x.",
		DEVICE, id->vendor, id->device );

	pbd = kmalloc(sizeof(*pbd), GFP_KERNEL);
	if(!pbd) {
		drv_err("%s: Failed to allocate memory.", DEVICE);
		return -ENOMEM;
	}
	if(pci_enable_device(dev)) {
		drv_err("%s: Failed to enable PCI device.", DEVICE);
		kfree(pbd);
		return -ENODEV;
	}
	if(pci_request_regions(dev, DEVICE)) {
		drv_err("%s: Failed to request PCI regions.", DEVICE);
		kfree(pbd);
		return -ENODEV;
	}
	if(dev_alloc(dev, pbd))
	{
		drv_err("%s: Failed to setup device.", DEVICE);
		kfree(pbd);
		pci_release_regions(dev);
		return -ENODEV;
	}
	pci_set_drvdata(dev, pbd);
	device_create_file(&dev->dev, &dev_attr_serial);
	pbd->minor = edredaq_register_dev( &dev->dev,
					&device_fops,
					(void *)pbd );
	kref_init(&pbd->kref);
	return 0;
}

static void __devexit dev_remove( struct pci_dev *dev )
{
	pBoardDevice_t pbd = (pBoardDevice_t)pci_get_drvdata(dev);
	dev_dealloc(pbd);
	edredaq_deregister_dev(pbd->minor);
	device_remove_file(&dev->dev, &dev_attr_serial);
	pci_release_regions(dev);
	pci_disable_device(dev);
	pci_set_drvdata(dev,NULL);
}

static int dev_suspend( struct pci_dev *dev, u32 state )
{
	return -ENODEV;
}

static int dev_resume( struct pci_dev *dev )
{
	return -ENODEV;
}

static int dev_enable_wake( struct pci_dev *dev, u32 state, int enable )
{
	return -ENODEV;
}

static struct pci_device_id pci_ids_tbl[] = {
	{ vendor: 0xea01l, device: 0x0052l,
	  subvendor: PCI_ANY_ID, subdevice: PCI_ANY_ID,
	  class: 0, class_mask: 0, driver_data: 0},
    	{ 0, }
};
MODULE_DEVICE_TABLE(pci,pci_ids_tbl);

static struct pci_driver pci_drv = {
	.name = DEVICE,
	.probe = dev_probe,
	.remove = __devexit_p(dev_remove),
	.suspend = dev_suspend,
	.resume = dev_resume,
	.enable_wake = dev_enable_wake,
	.id_table = pci_ids_tbl,
};

static ssize_t drv_show_version( struct device_driver *drv, char *buf )
{
	strcpy(buf, DRIVER_VERSION);
	return strlen(DRIVER_VERSION) + 1;
}

static DRIVER_ATTR(version, S_IRUGO, drv_show_version, NULL);

static int __init start_module(void)
{
	drv_info("%s (%s)", DRIVER_DESC, DRIVER_VERSION);
	if(pci_register_driver(&pci_drv)<0)
	{
		drv_err("%s: Failed to register driver with PCI sub-system.", DEVICE);
		return -ENODEV;
	}
	driver_create_file(&pci_drv.driver, &driver_attr_version);
	return EDRE_OK;
}

static void __exit stop_module(void)
{
	drv_info("%s: unloading...", DEVICE);
	driver_remove_file(&pci_drv.driver, &driver_attr_version);
	pci_unregister_driver(&pci_drv);
}

module_init(start_module);
module_exit(stop_module);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

/**
 * End of File
 */
