/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * buffer.h -- Buffer manager source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/
#include "stdinc.h"
#include "buffer.h"
#include "version.h"
#include "pci766inc.h"
 
void buffer_init( pBuffer_t pbuf, unsigned long BufferSize, void *RegSRam )
{
	spin_lock_init(&pbuf->lock);
	pbuf->regsram = RegSRam;
	pbuf->size = 0;
	pbuf->data = (unsigned short *)vmalloc(BufferSize * sizeof(unsigned short));
	if(pbuf->data) {
		pbuf->size = BufferSize;
		info( "%s: Allocated circular buffer of %ld kB.", DEVICE, (pbuf->size * sizeof(unsigned short)) / 1024 );
	}
	else
		err( "%s: Failed to allocate memory for D/A buffer.", DEVICE );
}

void buffer_uninit( pBuffer_t pbuf )
{
	if(pbuf->size)
		vfree(pbuf->data);
}

void buffer_reset( pBuffer_t pbuf )
{
	spin_lock(&pbuf->lock);
	pbuf->head = 0;
	pbuf->tail = 0;
	pbuf->underrun = 0;
	pbuf->wrap = 0;
	pbuf->samples = 0;
	spin_unlock(&pbuf->lock);
}
 
long buffer_getspace( pBuffer_t pbuf )
{
	long value;
	spin_lock(&pbuf->lock);
	value = pbuf->size - pbuf->samples;
	spin_unlock(&pbuf->lock);
	return value;
}

long buffer_adddata( pBuffer_t pbuf, unsigned char *buffer, unsigned long length )
{
	long p;
	unsigned long flags;
	unsigned long samples;
	long *uv = (long *)buffer;
	spin_lock_irqsave(&pbuf->lock, flags);
	
	samples = length / sizeof(long);
	
	for(p=0; p<samples; p++) {
		pbuf->data[pbuf->head++] = buffer_uvtobin(uv[p]);		
		
		if(pbuf->head == pbuf->size) {
			pbuf->wrap++;
			pbuf->head = 0;
		}
		
		if(++pbuf->samples == pbuf->size) {
			spin_unlock(&pbuf->lock);
			err( "%s: Buffer overrun.", DEVICE );
			return -1;
		}
	}
	spin_unlock_irqrestore(&pbuf->lock, flags);
	return 0;
}

long buffer_updatefifo( pBuffer_t pbuf )
{
	unsigned long flags, p;
	spin_lock_irqsave(&pbuf->lock, flags);

	for(p=0; p<pbuf->waterlevel; p++) {
		writel(pbuf->data[pbuf->tail++], pbuf->regsram);
		if(pbuf->tail == pbuf->size) pbuf->tail=0;
		pbuf->samples--;
	}
	spin_unlock_irqrestore(&pbuf->lock, flags);
	return 0;
}

long buffer_fillfifo( pBuffer_t pbuf )
{
	unsigned long flags, i = 0;
	spin_lock_irqsave(&pbuf->lock, flags);

	do {
		writel(pbuf->data[pbuf->tail++], pbuf->regsram);
		i++;
	} while(--pbuf->samples);
	
	pbuf->tail = 0;
	pbuf->head = 0;

	spin_unlock_irqrestore(&pbuf->lock, flags);
	return 0;
}


unsigned short buffer_uvtobin( long uv )
{	
	unsigned long bin;
	uv += 10000000;
	bin = uv / 305176;
	uv -= bin * 305176;
	bin *= 1000;
	bin += uv / 305;
	if(bin > 65535)
		bin = 65535;
	return bin;
}

long buffer_getstatus( pBuffer_t pbuf )
{
	long status;
	spin_lock(&pbuf->lock);
	status = pbuf->underrun;
	spin_unlock(&pbuf->lock);
	return status;
}

long buffer_testfifo(pBuffer_t pbuf)
{
	unsigned long count, val;
	for(count=0; count<65536; count++)
		writel(count, pbuf->regsram);
	readl(pbuf->regsram);
	for(count=1; count<65536; count++) {
		val = readl(pbuf->regsram) & 0xffff;
		if(val != count)
			return -1;
	}
	return 0;
}

/**
 * End of File
 */
