 /*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * ioctrl.c -- I/O Control Main source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/
#include "stdinc.h"
#include "version.h"
#include "pci766inc.h"
#include "ct.h"
#include "dac.h"
#include "buffer.h"
#include "isr.h"
#include "da8802.h"
#include "eeprom.h"
#include "pci766.h"
#include "ioctrl.h"
 
int ioctl_dioread( pBoardDevice_t pdev, unsigned long arg )
{
	unsigned long param[2], port, value;
	copy_from_user(param, (void *)arg, sizeof(unsigned long));
	port = param[0];
	if(port >= pdev->devinfo.diochan)
		return -EINVAL;
	value = readl(pdev->regmiscctrl) & (1<<port);
	if(value > 0)
		value = 0;
	else
		value = 1;
	param[0] = value;
	copy_to_user((void *)arg, param, sizeof(unsigned long));
	return 0;
}

int ioctl_dacfg( pBoardDevice_t pdev, unsigned long arg )
{
	pDAC_t pdac = &pdev->dac;
	pBuffer_t pbuf = &pdev->buffer;
	unsigned long param[5], chan, freq, clk, gate, cont;
	
	copy_from_user(param, (void *)arg, 5*sizeof(unsigned long));
	chan = param[0];
	freq = param[1];
	clk = param[2];
	gate = param[3];
	cont = param[4];
	
	if(dac_isbusy(pdac)) {
		err("%s: Device is busy.", DEVICE);
		return -EBUSY;
	}
	
	dac_fiforeset(pdac);
	buffer_reset(pbuf);
	dac_setalloc(pdac, chan);
	dac_configure(pdac, clk&0xff, (clk>>8) & 0xff, gate, freq, cont);
	return 0;
}

int ioctl_dacontrol( pBoardDevice_t pdev, unsigned long arg )
{
	pDAC_t pdac = &pdev->dac;
	unsigned long param[2], chan, command;
	
	copy_from_user(param, (void *)arg, 2*sizeof(unsigned long));
	chan = param[0];
	command = param[1];
	
	switch(command) {
	case DA_COMMAND_START:
		if(dac_isbusy(pdac))
			return -EBUSY;
		dac_setstartchan(pdac, chan);
		dac_enable(pdac);
		break;
	case DA_COMMAND_STOP:
		dac_disable(pdac);
		break;
	default:
		err( "%s: DAC command not supported", DEVICE);
		return -EINVAL;
	}
	return 0;
}

int ioctl_dawrite( pBoardDevice_t pdev, unsigned long arg )
{
	pDAC_t pdac = &pdev->dac;
	unsigned long param[2], port, bin;
	long value;
	
	copy_from_user(param,(void *)arg,2*sizeof(unsigned long));
	port=param[0];
	value=param[1];
	
	if(port >= pdev->devinfo.dacchan){	
		err("%s: Device does not have any D/A channels.", DEVICE);	
		return -EINVAL;
	}
	if(dac_isbusy(pdac)) {
		return -EBUSY;
	}
	if( (value > 10000000) || (value < -10000000) ) {
		err("%s: Voltage is out of range.", DEVICE);
		return -EINVAL;
	}
	
	bin = buffer_uvtobin(value);
	writel(bin, pdev->regdac0 + (port * 4));
	return 0;
}

int ioctl_ctcfg( pBoardDevice_t pdev, unsigned long arg )
{
	pCounter_t pct;
	unsigned long param[5], port, mode, type, source, gate;
	
	copy_from_user(param, (void *)arg, 5 * sizeof(unsigned long));
	port = param[0];
	mode = param[1];
	type = param[2];
	source = param[3];
	gate = param[4];
	
	if( (port > 1) ||
		(mode > 5) ||
		(source > 1) ||
		(gate>2) )
			return -EINVAL;
	pct = &pdev->ct[port];
	ct_configure(pct, mode, source, gate);
	return 0;
}

int ioctl_ctread( pBoardDevice_t pdev, unsigned long arg )
{
	pCounter_t pct;
	unsigned long param[1], port;
	
	copy_from_user(param, (void *)arg, sizeof(unsigned long));
	port=param[0];
	
	if(port > 1)
		return -EINVAL;
		
	pct = &pdev->ct[port];
	param[0] = ct_read(pct);
	copy_to_user((void *)arg,param,sizeof(unsigned long));
	return 0;
}

int ioctl_ctwrite( pBoardDevice_t pdev, unsigned long arg )
{
	pCounter_t pct;
	unsigned long param[2], port, value;
	
	copy_from_user(param, (void *)arg, 2 * sizeof(unsigned long));
	port = param[0];
	value = param[1];
	
	if(port > 1)
		return -EINVAL;
		
	pct = &pdev->ct[port];
	ct_write(pct, value);
	return 0;
}

int ioctl_ctsoftgate( pBoardDevice_t pdev, unsigned long arg )
{
	pCounter_t pct;
	unsigned long param[2], port, gate;
	
	copy_from_user(param, (void *)arg, 2 * sizeof(unsigned long));
	port = param[0];
	gate = param[1];
	
	if( (port > 1) ||
		(gate>2) )
			return -EINVAL;
			
	pct = &pdev->ct[port];
	ct_gate(pct, gate);
	return 0;
}

int ioctl_query(pBoardDevice_t pdev, unsigned long arg)
{
	unsigned long param[2];
	copy_from_user(param, (void *)arg, 2*sizeof(unsigned long));
	switch(param[0]) {
	case BRDSERIALNO:
		param[0] = pdev->devinfo.serialnumber;
		break;
	case BRDREV:
		param[0] = pdev->devinfo.revision;
		break;
	case BRDTYPE:
		param[0] = pdev->devinfo.boardtype;
		break;
	case BRDYEAR:
		param[0] = pdev->devinfo.year;
		break;
	case BRDMONTH:
		param[0] = pdev->devinfo.month;
		break;
	case BRDDAY:
		param[0] = pdev->devinfo.day;
		break;
	case BRDBUSTYPE:
		param[0] = PC_BUS_PCI;
	case DRVMAJOR:
		param[0] = (PCI766DRV_VERSION>>16)&0xff;
		break;
	case DRVMINOR:
		param[0] = (PCI766DRV_VERSION>>8)&0xff;
		break;
	case DRVBUILD:
		param[0] = PCI766DRV_VERSION&0xff;
		break;
	case ADNUMCHAN:
		param[0] = 0;
		break;
	case DANUMCHAN:
		param[0] = pdev->devinfo.dacchan;
		break;
	case CTNUM:
		param[0] = pdev->devinfo.cntchan;
		break;
	case DIONUMPORT:
		param[0] = pdev->devinfo.diochan;
		break;
	case DIOQRYPORT:
		param[0] = DIOIN;
		break;
	case DIOPORTWIDTH:
		param[0] = 1;
		break;
	case INTNUMSRC:
		param[0] = 0;
		break;
	default:
		param[0] = -1;
	}
	copy_to_user((void *)arg, param, sizeof(unsigned long));
	return 0;
}

/**
 * End of File
 */
