/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * pci766.c -- main source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/

#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
# error "This driver needs kernel 2.6.9 and later"
#endif
#include <linux/config.h>
#ifndef CONFIG_PCI
# error "This driver needs PCI support"
#endif
#include "stdinc.h"
#include "eeprom.h"
#include "ct.h"
#include "dac.h"
#include "eeprom.h"
#include "da8802.h"
#include "buffer.h"
#include "isr.h"
#include "pci766inc.h"
#include "pci766.h"
#include "ioctrl.h"
#include "version.h"

#define DRIVER_AUTHOR "Johan Russouw, johan@eagle.co.za & Jan Zwiegers, \
			jan@eagle.co.za"
#define DRIVER_DESC "PCI766 Interface Driver (C) 2004 - 2005 Eagle Technology"

static int dev_read( struct file *file, char *buf, size_t length,
			loff_t * offset )
{
	return -EINVAL;
}

static ssize_t dev_write( struct file *file, const char *buf, size_t length, 
			loff_t * offset )
{
	pBoardDevice_t pdev;
	pBuffer_t pbuf;
	pDAC_t pdac;
	unsigned long tocopy, notcopied, transfered, size;

	pdev = (pBoardDevice_t)file->private_data;
	pbuf = &pdev->buffer;
	pdac = &pdev->dac;
	size = sizeof(unsigned long);
	
	transfered = 0;
	do {
		tocopy = length;
		if(tocopy > (pdev->dblbuffer.size))
			tocopy = pdev->dblbuffer.size;
			
		notcopied = copy_from_user( (void *)pdev->dblbuffer.data,
					(void *)(buf + transfered),
					tocopy);
	
		if(notcopied >= 0) {
			transfered += (tocopy - notcopied);
			length -= (tocopy - notcopied);
			buffer_adddata( pbuf, pdev->dblbuffer.data, tocopy);
		}
		else {
			err( "%s: Data copy error occured (#%ld).", 
				DEVICE, notcopied );
			return -EINVAL;
		}		
	} while(length);
	
	if(dac_getmode(pdac) == DAC_MODE_LOOP)
		buffer_fillfifo(pbuf);
	
	return transfered;	
}

static void dev_kref_release( struct kref *kref )
{
	struct BoardDevice_t *dev = NULL;
	dev = container_of(kref, struct BoardDevice_t, kref);
	kfree(dev);
}

static int dev_open( struct inode *inode, struct file *pfile )
{
	pBoardDevice_t pdev;
	pdev = (pBoardDevice_t)pfile->private_data;
	spin_lock(&pdev->lock);
	spin_unlock(&pdev->lock);
	pfile->f_pos = 0;
	kref_get(&pdev->kref);
	return 0;
}

static int dev_release( struct inode *inode, struct file *pfile )
{
	pBoardDevice_t pdev = (pBoardDevice_t)pfile->private_data;
	kref_put(&pdev->kref, dev_kref_release);
	return 0;
}

static int dev_ioctl( struct inode *inode, struct file *pfile,
		      unsigned int cmd, unsigned long arg )
{
	pBoardDevice_t pdev = (pBoardDevice_t)pfile->private_data;
	int eerr = 0;
	int size = _IOC_SIZE(cmd);

	if(_IOC_TYPE(cmd) != DEV_IOC_MAGIC)
		return -EINVAL;
	if(_IOC_NR(cmd) > DEV_IOC_MAXNR)
		return -EINVAL;
	if(_IOC_DIR(cmd) & _IOC_READ)
		eerr = verify_area(VERIFY_WRITE, (void *)arg, size);
	else if(_IOC_DIR(cmd) & _IOC_WRITE)
		eerr = verify_area(VERIFY_READ, (void *)arg, size);
	if(eerr)
		return eerr;
    
	switch(cmd) {
	case DEV_IOC_QUERY:
		eerr = ioctl_query(pdev, arg);
		break;
	case DEV_IOC_DIOREAD:
		eerr = ioctl_dioread(pdev, arg);
		break;
	case DEV_IOC_DAWRITE:
		eerr = ioctl_dawrite(pdev, arg);
		break;
	case DEV_IOC_DACONTROL:
		eerr = ioctl_dacontrol(pdev, arg);
		break;
	case DEV_IOC_DACFG:
		eerr = ioctl_dacfg(pdev, arg);
		break;
	case DEV_IOC_CTREAD:
		eerr = ioctl_ctread(pdev, arg);
		break;
	case DEV_IOC_CTWRITE:
		eerr = ioctl_ctwrite(pdev, arg);
		break;
	case DEV_IOC_CTCFG:
		eerr = ioctl_ctcfg(pdev, arg);
		break;
	case DEV_IOC_CTSOFTGATE:
		eerr = ioctl_ctsoftgate(pdev, arg);
		break;
	default:
		eerr = -EINVAL;
	}
	return eerr;
}

static struct file_operations device_fops = {
	.owner = THIS_MODULE,
	.write = dev_write,
	.read = dev_read,
	.open = dev_open,
	.release = dev_release,
	.ioctl = dev_ioctl,
};

void dev_inittype( pBoardDevice_t pdev )
{
	switch(pdev->devinfo.boardtype) {
	case PCI766A8:
		pdev->devinfo.dacchan = 8;
		pdev->devinfo.diochan = 6;
		break;
	case PCI766A16:
		pdev->devinfo.dacchan = 16;
		pdev->devinfo.diochan = 6;
		break;
	case PCI766A24:
		pdev->devinfo.dacchan = 24;
		pdev->devinfo.diochan = 6;
		break;
	default:
		pdev->devinfo.dacchan = 0;
		pdev->devinfo.diochan = 0;
		break;
	}
}
void dev_readeeprom( pBoardDevice_t pdev)
{
	pEeprom_t pee = &pdev->eeprom;
	pCalDac_t pcal = &pdev->caldac;
	pdev->devinfo.serialnumber = ee_read(pee, 20)<<16;
	pdev->devinfo.serialnumber |= ee_read(pee, 21);
	pdev->devinfo.year = ee_read(pee, 22);
	pdev->devinfo.month = ee_read(pee, 23) & 0xff;
	pdev->devinfo.day = (ee_read(pee,23) >> 8) & 0xff;
	pdev->devinfo.boardtype = ee_read(pee, 24);
	pdev->devinfo.revision = ee_read(pee, 25);
		
	caldac_write(pcal, 0, ee_read(pee, 0));
	caldac_write(pcal, 1, ee_read(pee, 1));
	caldac_write(pcal, 2, ee_read(pee, 2));
	caldac_write(pcal, 3, ee_read(pee, 3));
}

int dev_hwinit(pBoardDevice_t pdev)
{
	irq_init( &pdev->isr,
				&pdev->buffer,
				pdev->ptrmem + REG_INTERRUPT_CONTROL,
				pdev->ptrmem + REG_INTERRUPT_MASK,
				pdev->ptrmem + REG_INTERRUPT_STATUS,
				pdev->ptrmem + REG_DMA_CNTRL );
	ee_init(&pdev->eeprom, pdev->ptrmem + REG_MISC_CONTROL);
	caldac_init(&pdev->caldac, pdev->ptrmem + REG_MISC_CONTROL);
	pdev->regmiscctrl = pdev->ptrmem + REG_MISC_CONTROL;
	pdev->regdac0 = pdev->ptrmem + REG_DAC0_PIO;
	dev_readeeprom(pdev);

	if( (pdev->devinfo.serialnumber < 1000000001) ||
		(pdev->devinfo.serialnumber == 0xffffffff)) {
		err( "%s: Invalid Serial Number (%ld).",
		     DEVICE, pdev->devinfo.serialnumber);
		return -ENODEV;
	}
	if( (pdev->devinfo.boardtype < PCI766A8) ||
		(pdev->devinfo.boardtype > PCI766A24)) {
		err( "%s: Invalid Board Type (%d).",
		     DEVICE, pdev->devinfo.boardtype);
		return -ENODEV;
	}

	info( "%s: Added new device (serial number #%ld, type #%d).",
	      DEVICE, pdev->devinfo.serialnumber, pdev->devinfo.boardtype);
	
	dev_inittype(pdev);
	buffer_init(&pdev->buffer, BUFFERSIZE, pdev->ptrmem + REG_SRAM);
	ct_init( &pdev->ct[0],
			pdev->ptrmem + REG_COUNTER0_CONFIG,
			pdev->ptrmem + REG_COUNTER0_MODE,
			pdev->ptrmem + REG_COUNTER0_COUNT );
	ct_init( &pdev->ct[1],
			pdev->ptrmem + REG_COUNTER1_CONFIG,
			pdev->ptrmem + REG_COUNTER1_MODE,
			pdev->ptrmem + REG_COUNTER1_COUNT);
	dac_init( &pdev->dac,
			&pdev->ct[0],
			pdev->ptrmem + REG_CHAN_LIST_CONFIG,
			pdev->ptrmem + REG_RESET_FIFO,
			pdev->ptrmem + REG_WATER_MARK);

//	irq_enable(&pdev->isr);
	pdev->dblbuffer.data = kmalloc(DOUBLE_BUFFER_SIZE, GFP_KERNEL);
	if(pdev->dblbuffer.data == NULL)
		pdev->dblbuffer.size = 0;
	else
		pdev->dblbuffer.size = DOUBLE_BUFFER_SIZE;
	info( "%s: Allocated double buffer of %d bytes.",
	      DEVICE, pdev->dblbuffer.size);
	
	if(buffer_testfifo(&pdev->buffer)) {
		err( "%s: Hardware FIFO test failed.", DEVICE);
		return -ENODEV;
	}
		
	return 0;
}

void dev_hwuninit(pBoardDevice_t pdev)
{

	dac_disable(&pdev->dac);
//	irq_disable(&pdev->isr);
	buffer_uninit(&pdev->buffer);
	if(pdev->dblbuffer.data)
		kfree(pdev->dblbuffer.data);
}

int dev_alloc( struct pci_dev *pcidev, pBoardDevice_t pdev )
{
	spin_lock_init(&pdev->lock);
	spin_lock(&pdev->lock);
	pdev->pci_device = pcidev;
	pdev->isr.pci_device = pcidev;
	
	pdev->ptrmem = ioremap_nocache( pci_resource_start(pcidev, 0),
					 pci_resource_len(pcidev, 0));
	if(!pdev->ptrmem) {
		spin_unlock(&pdev->lock);
		err("%s: Unable to remap BAR 0.", DEVICE);
		return -ENOMEM;
	}
	
	if(dev_hwinit(pdev)) {
		spin_unlock(&pdev->lock);
		err("%s: Failed to initialize hardware.", DEVICE);
		return -ENODEV;
	}
	spin_unlock(&pdev->lock);
	return 0;
}

void dev_dealloc(pBoardDevice_t pdev)
{
	spin_lock(&pdev->lock);
	dev_hwuninit(pdev);
	iounmap(pdev->ptrmem);
	spin_unlock(&pdev->lock);
}

static ssize_t dev_show_serial( struct device *device, char *buf )
{
	char s[20];
	pBoardDevice_t pbd = (pBoardDevice_t)device->driver_data;
	sprintf(s,"%ld\n", pbd->devinfo.serialnumber);
	strcpy(buf, s);
	return strlen(s) + 1;
}
static DEVICE_ATTR(serial, S_IRUGO, dev_show_serial, NULL);

static int __devinit dev_probe( struct pci_dev *dev,
				const struct pci_device_id *id )
{
	pBoardDevice_t pbd;

	info( "%s: Probing device: vendor = %x, device = %x.",
		DEVICE, id->vendor, id->device );

	pbd = kmalloc(sizeof(*pbd), GFP_KERNEL);
	if(!pbd) {
		err("%s: Failed to allocate memory.", DEVICE);
		return -ENOMEM;
	}
	if(pci_enable_device(dev)) {
		err("%s: Failed to enable PCI device.", DEVICE);
		kfree(pbd);
		return -ENODEV;
	}
	if(pci_request_regions(dev, DEVICE)) {
		err("%s: Failed to request PCI regions.", DEVICE);
		kfree(pbd);
		return -ENODEV;
	}
	if(dev_alloc(dev, pbd))
	{
		err("%s: Failed to setup device.", DEVICE);
		kfree(pbd);
		pci_release_regions(dev);
		return -ENODEV;
	}
	pci_set_drvdata(dev, pbd);
	device_create_file(&dev->dev, &dev_attr_serial);
	pbd->minor = edredaq_register_dev( &dev->dev,
					&device_fops,
					(void *)pbd );
	kref_init(&pbd->kref);
	return 0;
}

static void __devexit dev_remove( struct pci_dev *dev )
{
	pBoardDevice_t pbd = (pBoardDevice_t)pci_get_drvdata(dev);
	dev_dealloc(pbd);
	edredaq_deregister_dev(pbd->minor);
	device_remove_file(&dev->dev, &dev_attr_serial);
	pci_release_regions(dev);
	pci_disable_device(dev);
	pci_set_drvdata(dev,NULL);
}

static int dev_suspend( struct pci_dev *dev, u32 state )
{
	return -ENODEV;
}

static int dev_resume( struct pci_dev *dev )
{
	return -ENODEV;
}

static int dev_enable_wake( struct pci_dev *dev, u32 state, int enable )
{
	return -ENODEV;
}

static struct pci_device_id pci_ids_tbl[] = {
	{ vendor: 0xea01l, device: 0x0046l,
	  subvendor: PCI_ANY_ID, subdevice: PCI_ANY_ID,
	  class: 0, class_mask: 0, driver_data: 0},
    	{ 0, }
};
MODULE_DEVICE_TABLE(pci,pci_ids_tbl);

static struct pci_driver pci_drv = {
	.name = DEVICE,
	.probe = dev_probe,
	.remove = __devexit_p(dev_remove),
	.suspend = dev_suspend,
	.resume = dev_resume,
	.enable_wake = dev_enable_wake,
	.id_table = pci_ids_tbl,
};

static ssize_t drv_show_version( struct device_driver *drv, char *buf )
{
	strcpy(buf, DRIVER_VERSION);
	return strlen(DRIVER_VERSION) + 1;
}

static DRIVER_ATTR(version, S_IRUGO, drv_show_version, NULL);


static int __init start_module(void)
{
	info("%s (%s)", DRIVER_DESC, DRIVER_VERSION);
	if(pci_register_driver(&pci_drv)<0)
	{
		err("%s: Failed to register driver with PCI sub-system.", DEVICE);
		return -ENODEV;
	}
	driver_create_file(&pci_drv.driver, &driver_attr_version);
	return EDRE_OK;
}

static void __exit stop_module(void)
{
	info("%s: unloading...", DEVICE);
	driver_remove_file(&pci_drv.driver, &driver_attr_version);
	pci_unregister_driver(&pci_drv);
}

module_init(start_module);
module_exit(stop_module);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

/**
 * End of File
 */
