/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * usbupc.c -- USBuPc main source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/
#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
# error "This driver needs kernel 2.6.9 and later"
#endif
#include <linux/config.h>
#ifndef CONFIG_USB
# error "This driver needs USB support"
#endif
#include "stdinc.h"
#include "version.h"
#include "usbupc.h"
#include "usbio.h"
#include "deviceio.h"
#include "ioctl.h"
#include "buffer.h"
#include <edredaq.h>

/* module info */
#define DRIVER_AUTHOR "Jan Zwiegers, jan@eagle.co.za"
#define DRIVER_DESC "USBuPc Linux Interface Driver (C) 2004-2005"
#define MODULE_NAME "usbupcdrv"

/* global varables */
static struct file_operations usbupc_fops;
static struct usb_driver usbupc_driver;
static DECLARE_MUTEX (discon_sem);
static DECLARE_MUTEX (read_sem);
static DECLARE_MUTEX (ioctl_sem);
static DECLARE_MUTEX (write_sem);

/* file operations */
static loff_t usbupc_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}

static int usbupc_read(struct file *file, char *buf, size_t length, loff_t *
			offset)
{
	pUSBuPcDevice_t pdev;
	pADC_t padc;
	pDblBuffer_t pdblbuffer;
	unsigned long samples, tocopy, copied, transfered, size;
	int eerr = 0;

	pdev = (pUSBuPcDevice_t)file->private_data;
	padc = &pdev->adc;
	pdblbuffer = &padc->dblbuffer;
	size = sizeof(unsigned short);
	if(padc->datamode == ADC_DATA_MODE_VOLT)
		size = sizeof(long);

	samples = length / size;
	transfered = 0;
	do {
		tocopy = samples;
		if(tocopy > (pdblbuffer->size / size))
			tocopy = pdblbuffer->size / size;
		if(padc->datamode != ADC_DATA_MODE_VOLT)
			copied = usbupc_buf_getdataraw( &padc->buffer,
						  (unsigned short *)pdblbuffer->data, tocopy);
		else
			copied = usbupc_buf_getdata(padc, (long *)pdblbuffer->data, tocopy);
		samples -= copied;
		eerr = tocopy - copied;
		do {
			tocopy = copied;
			copied = copy_to_user( (void *)(buf + transfered * size),
									(void *)pdblbuffer->data, tocopy * size);
			if(copied)
				copied /= size;
			transfered += tocopy - copied;
		}while(copied > 0);
	}while(samples > 0 || eerr);
	transfered -= samples;
	if(eerr)
		return -EINVAL;
	return transfered * size;
}

static int usbupc_write(struct file *file, const char *buf, size_t length,
			loff_t * offset)
{
	return -ESPIPE;
}

static void kref_release(struct kref *kref)
{
	struct USBuPcDevice_t *dev=NULL;
	dev = container_of(kref, struct USBuPcDevice_t, kref);
	kfree(dev);
}

static int usbupc_release(struct inode *inode, struct file *file)
{
	pUSBuPcDevice_t dev = (pUSBuPcDevice_t)file->private_data;
	if(!dev)
		return -ENOMEM;
	usbupc_int_stop(dev);
	usbupc_isoxfr_stop(dev);
	kref_put(&dev->kref, kref_release);
	return 0;
}

static int usbupc_open(struct inode *inode, struct file *file)
{
	pUSBuPcDevice_t dev;
	down(&discon_sem);
	dev = (pUSBuPcDevice_t)file->private_data;
	file->f_pos = 0;
	file->private_data = dev;
	kref_get(&dev->kref);
	up(&discon_sem);
	return 0;
}

static int usbupc_ioctl(struct inode *inode, struct file *file, unsigned int
			cmd, unsigned long arg)
{
	pUSBuPcDevice_t dev = (pUSBuPcDevice_t)file->private_data;
	int eerr = 0;
	int size = _IOC_SIZE(cmd);

	if(_IOC_TYPE(cmd) != DEV_IOC_MAGIC)
		return -EINVAL;
	if(_IOC_NR(cmd) > DEV_IOC_MAXNR)
		return -EINVAL;
	if(_IOC_DIR(cmd) & _IOC_READ)
		eerr = verify_area(VERIFY_WRITE, (void *)arg, size);
	else if(_IOC_DIR(cmd)&_IOC_WRITE)
		eerr = verify_area(VERIFY_READ, (void *)arg, size);
	if(eerr)
		return eerr;

	if(!dev)
		return -ENOMEM;

	switch(cmd) {
	case DEV_IOC_QUERY:
		eerr = usbupc_ioctl_query(dev, arg);
		break;
	case DEV_IOC_DIOWRITE:
		eerr = usbupc_dio_write(dev, arg);
		break;
	case DEV_IOC_DIOREAD:
		eerr = usbupc_dio_read(dev, arg);
		break;
	case DEV_IOC_CTREAD:
		eerr = usbupc_cnt_read(dev, arg);
		break;
	case DEV_IOC_CTWRITE:
		eerr = usbupc_cnt_write(dev, arg);
		break;
	case DEV_IOC_CTCFG:
		eerr = usbupc_cnt_config(dev, arg);
		break;
	case DEV_IOC_CTSOFTGATE:
		eerr = usbupc_cnt_softgate(dev, arg);
		break;
	case DEV_IOC_INTCFG:
		eerr = usbupc_int_config(dev, arg);
		break;
	case DEV_IOC_INTEN:
		eerr = usbupc_int_enable(dev, arg);
		break;
	case DEV_IOC_INTDIS:
		eerr = usbupc_int_disable(dev, arg);
		break;
	case DEV_IOC_INTWAIT:
		eerr = usbupc_int_wait(dev, arg);
		break;
	case DEV_IOC_INTRELEASE:
		eerr = usbupc_int_release(dev, arg);
		break;
	case DEV_IOC_DAWRITE:
		eerr = usbupc_dac_write(dev, arg);
		break;
	case DEV_IOC_ADSINGLE:
		eerr = usbupc_adc_single(dev, arg);
		break;
	case DEV_IOC_ADCFG:
		eerr = usbupc_adc_config(dev, arg);
		break;
	case DEV_IOC_ADCLKSRC:
		eerr = usbupc_adc_clksrc(dev, arg);
		break;
	case DEV_IOC_ADCHLST:
		eerr = usbupc_adc_chlst(dev, arg);
		break;
	case DEV_IOC_ADFREQ:
		eerr = usbupc_adc_freq(dev, arg);
		break;
	case DEV_IOC_ADSTART:
		eerr = usbupc_adc_start(dev, arg);
		break;
	case DEV_IOC_ADSTOP:
		eerr = usbupc_adc_stop(dev, arg);
		break;
	default:
		eerr = -EINVAL;
	}
	return eerr;
}

static ssize_t dev_show_serial(struct device *device, char *buf)
{
	char s[20];
	pUSBuPcDevice_t dev = (pUSBuPcDevice_t)device->driver_data;
	sprintf(s,"%ld\n",dev->eeinfo.serialnumber);
	strcpy(buf,s);
	return strlen(s) + 1;
}
static DEVICE_ATTR(serial, S_IRUGO, dev_show_serial, NULL);

static struct file_operations usbupc_fops = {
	.owner = THIS_MODULE,
	.open = usbupc_open,
	.release = usbupc_release,
	.ioctl = usbupc_ioctl,
	.read = usbupc_read,
	.write = usbupc_write,
	.llseek = usbupc_llseek,
};

static int usbupc_probe(struct usb_interface *interface,
			const struct usb_device_id *id)
{
	pUSBuPcDevice_t dev;

	drv_info("usbupc: usbupc_probe - vendor id 0x%x, device id 0x%x",
	    id->idVendor, id->idProduct);

	dev = kmalloc(sizeof(*dev), GFP_KERNEL);
	if(dev == NULL) {
		drv_err("usbupc: Out of memory");
		return -ENOMEM;
	}

	spin_lock_init(&dev->spl_bulk);
	spin_lock_init(&dev->adc.lock);
	spin_lock_init(&dev->usbint.lock);
	spin_lock_init(&dev->lock);
	spin_lock_init(&dev->adc.buffer.lock);
	spin_lock_init(&dev->iso_context.lock);
	dev->usbint.datasz=64;
        dev->usbint.interval=1;

	if(usbupc_init_endpoints(dev, interface)) {
		drv_err("usbupc_probe: failed to initialize usb endpoint.");
		goto usbupc_probe_error;
	}

	if(usbupc_get_deviceinfo(dev)!=EOK) {
		drv_err("usbupc_probe: failed to retrieve device info.");
		goto usbupc_probe_error;
	}
	usbupc_set_devicetype(dev);
	usbupc_init_device(dev);
	usb_set_intfdata(interface, dev);

	device_create_file(&interface->dev, &dev_attr_serial);

	dev->minor = edredaq_register_dev( &interface->dev,
					    &usbupc_fops,
					    (void *)dev );
	kref_init(&dev->kref);
	return 0;

usbupc_probe_error:
	kref_put(&dev->kref, kref_release);
	return -ENOMEM;
}

static void usbupc_disconnect (struct usb_interface *interface)
{
	pUSBuPcDevice_t dev;
	lock_kernel();
	dev = (pUSBuPcDevice_t)usb_get_intfdata(interface);
	device_remove_file(&interface->dev, &dev_attr_serial);
	edredaq_deregister_dev(dev->minor);
	usbupc_uninit_device(dev);
	usb_set_intfdata(interface, NULL);
	unlock_kernel();
	kref_put(&dev->kref, kref_release);
	drv_info("usbupc: disconnecting device #%ld.", dev->eeinfo.serialnumber);
}


static struct usb_device_id usbupc_ids_tbl[] = {
	{ USB_DEVICE(0xea01l, 0x0001l) }, /*USB24A*/
	{ USB_DEVICE(0xea01l, 0x0002l) }, /*USB24C*/
	{ USB_DEVICE(0xea01l, 0x0003l) }, /*USB48A*/
	{ USB_DEVICE(0xea01l, 0x0004l) }, /*USB48C*/
	{ USB_DEVICE(0xea01l, 0x0005l) }, /*USB72A*/
	{ USB_DEVICE(0xea01l, 0x0006l) }, /*USB72C*/
	{ USB_DEVICE(0xea01l, 0x0007l) }, /*USB96A*/
	{ USB_DEVICE(0xea01l, 0x0008l) }, /*USB96C*/
	{ USB_DEVICE(0xea01l, 0x0009l) }, /*USB120A*/
	{ USB_DEVICE(0xea01l, 0x000Al) }, /*USB120C*/
	{ USB_DEVICE(0xea01l, 0x000Bl) }, /*USB62-16*/
	{ USB_DEVICE(0xea01l, 0x000Cl) }, /*USB62-32*/
	{ USB_DEVICE(0xea01l, 0x000Dl) }, /*USB63-16*/
	{ USB_DEVICE(0xea01l, 0x000El) }, /*USB63-32*/
	{ USB_DEVICE(0xea01l, 0x000Fl) }, /*USB69-16*/
	{ USB_DEVICE(0xea01l, 0x0010l) }, /*USB69-32*/	
	{ USB_DEVICE(0xea01l, 0x0013l) }, /*USB26A16*/
	{ USB_DEVICE(0xea01l, 0x0014l) }, /*USB26A32*/
	{ USB_DEVICE(0xea01l, 0x0015l) }, /*USB30A16*/
	{ USB_DEVICE(0xea01l, 0x0016l) }, /*USB30A32*/
	{ USB_DEVICE(0xea01l, 0x0017l) }, /*USB26B16*/
	{ USB_DEVICE(0xea01l, 0x0018l) }, /*USB26B32*/
	{ USB_DEVICE(0xea01l, 0x0019l) }, /*USB30B16*/
	{ USB_DEVICE(0xea01l, 0x001Al) }, /*USB30B32*/
	{ USB_DEVICE(0xea01l, 0x001Bl) }, /*USB26C16*/
	{ USB_DEVICE(0xea01l, 0x001Cl) }, /*USB26C32*/
	{ USB_DEVICE(0xea01l, 0x001Dl) }, /*USB30C16*/
	{ USB_DEVICE(0xea01l, 0x001El) }, /*USB30C32*/
	{ USB_DEVICE(0xea01l, 0x0020l) }, /*USB73T8*/
	{ USB_DEVICE(0xea01l, 0x0021l) }, /*USB73T16*/
	{ USB_DEVICE(0xea01l, 0x0022l) }, /*USB73T32*/
	{ USB_DEVICE(0xea01l, 0x0023l) }, /*USB73R8*/
	{ USB_DEVICE(0xea01l, 0x0024l) }, /*USB73R16*/
	{ USB_DEVICE(0xea01l, 0x0025l) }, /*USB73R32*/
	{ USB_DEVICE(0xea01l, 0x0030l) }, /*RUGGEDDAQ*/	
	{ } /*nothing*/
};

MODULE_DEVICE_TABLE(usb,usbupc_ids_tbl);

static struct usb_driver usbupc_driver = {
	.owner = THIS_MODULE,
	.name = MODULE_NAME,
	.probe = usbupc_probe,
	.disconnect = usbupc_disconnect,
	.id_table = usbupc_ids_tbl,
};

static ssize_t drv_show_version(struct device_driver *drv, char *buf)
{
	strcpy(buf, DRIVER_VERSION);
	return strlen(DRIVER_VERSION) + 1;
}

static DRIVER_ATTR(version, S_IRUGO, drv_show_version, NULL);

static int __init usbupc_init(void)
{
	drv_info(DRIVER_DESC " : " DRIVER_VERSION);
	if(usb_register(&usbupc_driver)) {
		drv_err("usbupc_init: failed to register driver with USB sub-system.");
		return -ENODEV;
	}
	driver_create_file(&usbupc_driver.driver, &driver_attr_version);
	return 0;
}

static void __exit usbupc_cleanup(void)
{	
	driver_remove_file(&usbupc_driver.driver, &driver_attr_version);
	usb_deregister(&usbupc_driver);
}

module_init(usbupc_init);
module_exit(usbupc_cleanup);
MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

/**
 * End of File
 */
