/*******************************************************************************
 *                 Eagle Technology - Cape Town, South Africa
 *             Copyright (C) 2004 - 2005 - All Rights Reserved!
 *                 www.eagledaq.com - eagle@eagle.co.za
 *******************************************************************************
 *
 * usbio.c -- USB I/O source file
 *
 *******************************************************************************
 *                              LICENSE AGREEMENT
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ******************************************************************************/
#include "stdinc.h"
#include "usbupc.h"
#include "buffer.h"
#include "usbio.h"

int usbupc_init_endpoints( pUSBuPcDevice_t dev, struct usb_interface *interface
)
{
	struct usb_device *ud;
	struct usb_host_interface *if_desc;
	struct usb_endpoint_descriptor *ep_desc;
	int i;

	dev->usbdev = usb_get_dev(interface_to_usbdev(interface));
	ud = dev->usbdev;
	dev->usbif = interface;
	if_desc = interface->cur_altsetting;

	dev->iso_ep=0;
	dev->bulk_ep=0;	

	for(i=0; i<if_desc->desc.bNumEndpoints; i++) {
		ep_desc = &if_desc->endpoint[i].desc;
		if( !dev->bulk_ep &&
		    (ep_desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
		    == USB_ENDPOINT_XFER_BULK) {
			dev->bulk_ep = ep_desc->bEndpointAddress&0xf;
		}
		else if( !dev->iso_ep &&
			(ep_desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
			== USB_ENDPOINT_XFER_ISOC) {
			dev->iso_ep = ep_desc->bEndpointAddress&0xf;
			dev->iso_ep_size = ep_desc->wMaxPacketSize;
			dev->iso_interval = ep_desc->bInterval;
		}
	}

	for(i=0; i<ISO_NUM_URB; i++) {
		dev->adc.isourb[i].allocated = 0;
	}

	if(dev->iso_ep)
		drv_info("usbupclite: Found ISO endpoint on %d. max_packet = %d.",
			dev->iso_ep,
			dev->iso_ep_size);
	if(dev->bulk_ep)
		drv_info("usbupclite: Found BULK endpoint on %d.", dev->bulk_ep);
	return EOK;
}

int usbupc_bulkxfr( pUSBuPcDevice_t dev, int dir, unsigned char *buffer,
		    int *size )
{
	int ret;
	unsigned int pipe;
	int ln;
	if(dir == PACKET_DIR_OUT)
		pipe = usb_sndbulkpipe(dev->usbdev, dev->bulk_ep);
	else
		pipe = usb_rcvbulkpipe(dev->usbdev, dev->bulk_ep);
	ret = usb_bulk_msg(dev->usbdev, pipe, buffer, *size, &ln, HZ*10);
	if(ret < EOK)
		drv_err("usbupclite: usb_bulk_msg failed(%d)",ret);
	if(ret == -EPIPE) {
		drv_err("usbupclite: usb_bulk_msg stalled");
		if(usb_clear_halt(dev->usbdev,usb_pipeendpoint(pipe)))
			drv_err("usbupclite: failed to clear stalled endpoint");
	}
	*size=ln;
	return ret;
}

int usbupc_bulkpacket( pUSBuPcDevice_t dev, unsigned char *buffer,
		       unsigned int size )
{
	int ret;
	int sz;
	sz = size;
	spin_lock(&dev->spl_bulk);
	ret = usbupc_bulkxfr(dev, PACKET_DIR_OUT, buffer, &sz);
	if(ret < EOK) {
		drv_err("usbupclite: bulk write failed");
		spin_unlock(&dev->spl_bulk);
		return ret;
	}
	sz = size;
	memset(buffer, 0, size);
	ret = usbupc_bulkxfr(dev, PACKET_DIR_IN, buffer, &sz);
	if(ret < EOK)
		drv_err("usbupclite: bulk read failed");
	spin_unlock(&dev->spl_bulk);
	return ret;
}

void usbupc_isoxfr_dealloc( pIsoUrb_t isourb )
{
	if(isourb->allocated) {
		if(isourb->urb->transfer_buffer) {
			usb_buffer_free ( isourb->urb->dev,
					isourb->urb->transfer_buffer_length,
					isourb->urb->transfer_buffer,
					isourb->urb->transfer_dma);

		}
		usb_free_urb(isourb->urb);
	}
	isourb->context->pending--;
	isourb->allocated=0;
}

static void usbupc_isoxfr_complete( struct urb *urb, struct pt_regs *regs )
{
	pIsoUrb_t isourb = (pIsoUrb_t)urb->context;
	pIsoContext_t ctx = isourb->context;
	void *buf = urb->transfer_buffer;
	int i, len, eerr=-1;	

	spin_lock(&ctx->lock);

	if(urb->error_count > 0 )
		ctx->errors += urb->error_count;

	if( (urb->status != -ENOENT) &&
	    (urb->status != -ENODEV) &&
	    (ctx->busy) ) {
		for(i=0; i<urb->number_of_packets; i++) {
			if((!urb->iso_frame_desc[i].status)) {
				len = urb->iso_frame_desc[i].actual_length;
				usbupc_buf_adddata(isourb->pbuffer,
				  (void*)(buf + urb->iso_frame_desc[i].offset), len);
			}
		}
		eerr = usb_submit_urb (urb, GFP_ATOMIC);
	}
	
	if(eerr) {
		usbupc_isoxfr_dealloc(isourb);
		if(ctx->pending == 0) {	
			ctx->busy = 0;
			complete(&ctx->done);
		}
	}

	spin_unlock(&ctx->lock);
	return;
}


int usbupc_isoxfr_alloc( struct usb_device *usbdev, pIsoUrb_t isourb )
{
	unsigned int pipe = usb_rcvisocpipe(usbdev, isourb->endpoint);
	int maxp;
	int pipesize = usb_maxpacket(usbdev, pipe, usb_pipeout(pipe));
	int packets = ISO_PIPESIZE / pipesize;
	int buffer_length = packets * pipesize;
	int i;

	maxp = 0x7ff & pipesize;
	maxp *= 1 + (0x3 & (pipesize >> 11));
	packets = (buffer_length + maxp - 1) / maxp;


	isourb->urb = usb_alloc_urb(packets, SLAB_KERNEL);
	if(!isourb->urb) {
		drv_err("usbupc_isoxfr_init(#%d): usb_alloc_urb == NULL.",
			isourb->index);
		return -ENOMEM;
	}

	isourb->urb->dev = usbdev;
	isourb->urb->pipe = pipe;
	isourb->urb->number_of_packets = packets;
	isourb->urb->transfer_buffer_length = buffer_length;
	isourb->urb->transfer_buffer = 
		usb_buffer_alloc( usbdev, buffer_length, SLAB_KERNEL,
				  &isourb->urb->transfer_dma);
	if (!isourb->urb->transfer_buffer) {
		usb_free_urb (isourb->urb);
		return -ENOMEM;
	}
	memset (isourb->urb->transfer_buffer, 0, buffer_length);

	for (i = 0; i < packets; i++) {
		isourb->urb->iso_frame_desc[i].length = pipesize;
		isourb->urb->iso_frame_desc[i].offset = pipesize * i;
	}

	isourb->urb->complete = usbupc_isoxfr_complete;
	isourb->urb->context = isourb;
	isourb->urb->interval = 1 << (isourb->interval - 1);
	isourb->urb->transfer_flags = URB_ISO_ASAP | URB_NO_TRANSFER_DMA_MAP;	
	spin_lock_init(&isourb->urb->lock);
	isourb->pipesize = pipesize;
	isourb->allocated = 1;
	isourb->context->pending++;
	return EDRE_OK;
}

int usbupc_isoxfr_start( pUSBuPcDevice_t dev )
{
	int eerr;

	if(dev->iso_context.busy)
		return -EBUSY;

	dev->iso_context.errors = 0;
	dev->iso_context.pending = 0;
	init_completion(&dev->iso_context.done);

	dev->adc.isourb[0].index = 0;
	dev->adc.isourb[0].allocated = 0;
	dev->adc.isourb[0].endpoint = dev->iso_ep;
	dev->adc.isourb[0].interval = dev->iso_interval;
	dev->adc.isourb[0].context = &dev->iso_context;

	if( usbupc_isoxfr_alloc(dev->usbdev,&dev->adc.isourb[0]) != EDRE_OK) {
		usbupc_isoxfr_dealloc(&dev->adc.isourb[0]);
		drv_err("usbupclite: Failed to allocate iso URB #0");
		return -ENOMEM;
	}

	dev->adc.isourb[1].index = 1;
	dev->adc.isourb[1].allocated = 0;
	dev->adc.isourb[1].endpoint = dev->iso_ep;
	dev->adc.isourb[1].interval = dev->iso_interval;
	dev->adc.isourb[1].context = &dev->iso_context;

	if( usbupc_isoxfr_alloc(dev->usbdev,&dev->adc.isourb[1]) != EDRE_OK) {
		drv_err("usbupclite: Failed to allocate iso endpoint 1");
		usbupc_isoxfr_dealloc(&dev->adc.isourb[0]);
		usbupc_isoxfr_dealloc(&dev->adc.isourb[1]);
		return -ENOMEM;
	}

	spin_lock_irq (&dev->iso_context.lock);
	eerr = usb_submit_urb(dev->adc.isourb[0].urb, GFP_ATOMIC);
	if(eerr) {
		drv_err("usbupclite: Isochronous #0 URB submit failed (#%d).",eerr);
		usbupc_isoxfr_dealloc(&dev->adc.isourb[0]);
		spin_unlock_irq (&dev->iso_context.lock);
		return -ENOMEM;
	}

	eerr = usb_submit_urb(dev->adc.isourb[1].urb, GFP_ATOMIC);
	if(eerr) {
		drv_err("usbupclite: Isochronous #1 URB submit failed (#%d).",eerr);
		usbupc_isoxfr_dealloc(&dev->adc.isourb[0]);
		usbupc_isoxfr_dealloc(&dev->adc.isourb[1]);
		spin_unlock_irq (&dev->iso_context.lock);
		return -ENOMEM;
	}

	dev->iso_context.busy = 1;
	spin_unlock_irq (&dev->iso_context.lock);

	return EDRE_OK;
}

void usbupc_isoxfr_stop( pUSBuPcDevice_t dev )
{
	unsigned char busy = dev->iso_context.busy;
	spin_lock_irq (&dev->iso_context.lock);
	dev->iso_context.busy = 0;
	spin_unlock_irq (&dev->iso_context.lock);
	if(busy)
		wait_for_completion (&dev->iso_context.done);
}

/**
 * End of File
 */
